using System;
using System.Globalization;
using System.Runtime.InteropServices;
using Microsoft.Win32;

namespace Crankedup.Support.TimeZoneMgnt
{
	/// <summary>
	/// A structure (similiar to the DateTime structure) that is used for 
	/// performing calculations of the differences it time between various
	/// time zones.  This is the equivalent of the SYSTEMTIME structure 
	/// in the Windows API.
	/// </summary>
	[StructLayout(LayoutKind.Sequential)]
	public struct SystemTime
	{
		public short Year;
		public short Month;
		public short DayOfWeek;
		public short Day;
		public short Hour;
		public short Minute;
		public short Second;
		public short Millisecond;

		#region overrides // Added AW 2002-11-25
		public override bool Equals(object obj) // Added AW 2002-11-25
		{
			//try
			//{
			if (obj == null) return false; 
			if (this.GetType() != obj.GetType()) return false;

			SystemTime theST = (SystemTime)obj;
			if(
				(this.Year == theST.Year) && 
				(this.Month == theST.Month) && 
				(this.DayOfWeek == theST.DayOfWeek) && 
				(this.Day == theST.Day) && 
				(this.Hour == theST.Hour) && 
				(this.Minute == theST.Minute) && 
				(this.Second == theST.Second) && 
				(this.Millisecond == theST.Millisecond)
				)
			{
				return true;
			}
			else
			{
				return false;
			}
			//}
			//catch
			//{
			//	return false;
			//}
		}

		public override int GetHashCode() // Added AW 2002-11-25
		{
			DateTime theDateTime = this.ToDateTime();
			return theDateTime.GetHashCode();
		}

		public static bool operator==(SystemTime st1, SystemTime st2) // Added AW 2002-11-25
		{
			return(st1.Equals(st2));
		}

		public static bool operator!=(SystemTime st1, SystemTime st2) // Added AW 2002-11-25
		{
			return(!st1.Equals(st2));
		}

		#endregion

		#region external functions
		[DllImport("kernel32.dll", ExactSpelling=true)]
		internal static extern void GetLocalTime(out SystemTime localTime);
		[DllImport("kernel32.dll", ExactSpelling=true)]
		internal static extern void GetSystemTime(out SystemTime localTime);
		#endregion

		#region public instance methods

		/// <summary>
		/// Converts the SystemTime structure to a System.DateTime structure
		/// </summary>
		/// <returns></returns>
		public DateTime ToDateTime()
		{
			//construct a DateTime object from the SystemTime object that was passed in
			return SystemTime.ToDateTime(this);
		}
		
		#endregion

		#region static time retrieval functions

		/// <summary>
		/// Get the current time in Utc Format
		/// </summary>
		public static SystemTime UtcNow
		{ 
			get
			{				
				//use the GetSystemTime function to return the current UTC time
				SystemTime sysTime;
				SystemTime.GetSystemTime(out sysTime);
				return sysTime;
			}
		}
		/// <summary>
		/// Get the current time in local format
		/// </summary>
		public static SystemTime Now
		{
			get
			{
				//use the GetLocalTime function to return the current local time
				SystemTime sysTime;
				SystemTime.GetLocalTime(out sysTime);
				return sysTime;
			}
		}

		#endregion

		#region static conversion function

		/// <summary>
		/// Converts a SystemTime structure to a System.DateTime structure
		/// </summary>
		/// <param name="time">The SystemTime structure to be converted</param>
		/// <returns>A System.DateTime structure that was created based on
		/// the value of the SystemTime structure.</returns>
		public static DateTime ToDateTime(SystemTime time)
		{
			//construct a DateTime object from the SystemTime object that was passed in
			return new DateTime(time.Year, time.Month, time.Day, time.Hour, 
				time.Minute, time.Second, time.Millisecond);
		}
		/// <summary>
		/// Converts a System.DateTime structure to a SystemTime structure
		/// </summary>
		/// <param name="time">System.DateTime structure to be converted</param>
		/// <returns>A SystemTime structure that was created based on
		/// the value of the System.DateTime structure.</returns>
		public static SystemTime FromDateTime(DateTime time)
		{ 
			//create the SystemTime object and populate it's values from the 
			//DateTime object that was passed in
			SystemTime sysTime;
			sysTime.Day = (short)time.Day;
			sysTime.DayOfWeek = (short)time.DayOfWeek;
			sysTime.Hour = (short)time.Hour;
			sysTime.Millisecond = (short)time.Millisecond;
			sysTime.Minute = (short)time.Minute;
			sysTime.Month = (short)time.Month;
			sysTime.Second = (short)time.Second;
			sysTime.Year = (short)time.Year;
			return sysTime;
		}

		#endregion
	}
}
