using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace Expl.Itinerary {
   /// <summary>
   /// Interface to HighTower data access layer
   /// </summary>
   public interface IItinerary : IDisposable {
      /// <summary>
      /// Collection of schedule definitions
      /// </summary>
      IScheduleCollection Schedules { get; }

      /// <summary>
      /// Collection of scheduled job instances
      /// </summary>
      IJobCollection Jobs { get; }

      /// <summary>
      /// Collection of registered tasks
      /// </summary>
      ITaskCollection Tasks { get; }
   }

   /// <summary>
   /// Interface to schedule data access layer
   /// Handle CRUD operations.
   /// </summary>
   public interface IScheduleCollection : IDisposable, IEnumerable<IScheduleDefinition> {
      /// <summary>
      /// Get schedule by name
      /// </summary>
      /// <param name="Name">Name</param>
      /// <returns>IScheduleInfo object</returns>
      /// 
      IScheduleDefinition this[string name] { get; }

      /// <summary>
      /// Check if a schedule exists
      /// </summary>
      /// <param name="name">Name</param>
      /// <returns>True if exists</returns>
      bool Exists(string name);

      /// <summary>
      /// Get count of Schedule objects
      /// </summary>
      int Count { get; }

      /// <summary>
      /// Add a new schedule
      /// </summary>
      /// <param name="Name">Name</param>
      /// <param name="Description">Description</param>
      /// <param name="TDL">TDL</param>
      /// <returns>IScheduleInfo object</returns>
      IScheduleDefinition Add(string Name, string Description, string TDL);

      /// <summary>
      /// Rename schedule
      /// </summary>
      /// <param name="OldName">Old schedule name</param>
      /// <param name="NewName">New schedule name</param>
      void Rename(string OldName, string NewName);

      /// <summary>
      /// Delete a schedule
      /// </summary>
      /// <param name="Name">Name</param>
      void Delete(string Name);
   }

   /// <summary>
   /// Schedule definition entity interface.
   /// Parses TDL to create underlying ISchedule object.
   /// </summary>
   public interface IScheduleDefinition : IDisposable {
      /// <summary>
      /// Get name
      /// </summary>
      string Name { get; }

      /// <summary>
      /// Get/set description
      /// </summary>
      string Description { get; set; }

      /// <summary>
      /// Get TDL expression
      /// </summary>
      string TDL { get; set; }

      /// <summary>
      /// Test TDL expression for validity
      /// </summary>
      bool IsValid();

      /// <summary>
      /// Get ISchedule object derived from TDL
      /// </summary>
      ISchedule GetSchedule();

      /// <summary>
      /// Get modified flag of entity
      /// </summary>
      bool IsModified();

      /// <summary>
      /// Refresh data from backend store
      /// </summary>
      void Refresh();

      /// <summary>
      /// Commit changes to backend store
      /// </summary>
      void Commit();
   }

   public interface IJobCollection : IDisposable, IEnumerable<IJobDefinition> {
      /// <summary>
      /// Job added event
      /// </summary>
      event JobEventHandler JobAdded;

      /// <summary>
      /// Job deleted event
      /// </summary>
      event JobEventHandler JobDeleted;

      /// <summary>
      /// Job changed event
      /// </summary>
      event JobEventHandler JobChanged;

      /// <summary>
      /// Get job by ID
      /// </summary>
      /// <param name="ID">Job ID</param>
      /// <returns>IJobDefinition object</returns>
      IJobDefinition this[Guid ID] { get; }

      /// <summary>
      /// Get count of Job objects
      /// </summary>
      int Count { get; }

      /// <summary>
      /// Check if a job exists
      /// </summary>
      /// <param name="name">Job ID</param>
      /// <returns>True if exists</returns>
      bool Exists(Guid ID);

      /// <summary>
      /// Add a job
      /// </summary>
      /// <param name="Name">Job name</param>
      /// <param name="Description">Job description</param>
      /// <param name="TDL">Job's associated TDL script</param>
      /// <param name="TaskName">Job's associated task name</param>
      /// <param name="Parameters">Task parameters</param>
      /// <returns>IJobDefinition object</returns>
      IJobDefinition Add(string Name, string Description, string TDL, string TaskName, ITaskParameters Parameters, bool Enabled);

      /// <summary>
      /// Delete a job by ID
      /// </summary>
      /// <param name="JobID">Job ID to delete</param>
      void Delete(Guid JobID);
   }

   public delegate void JobEventHandler(object Sender, JobEventArgs Args);

   public class JobEventArgs : EventArgs {
      private IJobDefinition _JobEntity;

      public JobEventArgs(IJobDefinition JobEntity)
         : base() {
         _JobEntity = JobEntity;
      }

      public IJobDefinition Job {
         get { return _JobEntity; }
      }
   }

   /// <summary>
   /// Job definition entity object interface
   /// </summary>
   public interface IJobDefinition : IDisposable {
      /// <summary>
      /// Get job ID
      /// </summary>
      Guid ID { get; }

      /// <summary>
      /// Get/set job name
      /// </summary>
      string Name { get; set; }

      /// <summary>
      /// Get/set job description
      /// </summary>
      string Description { get; set; }

      /// <summary>
      /// Get/set associated TDL
      /// </summary>
      string TDL { get; set; }

      /// <summary>
      /// Get/set associated task name
      /// </summary>
      string TaskName { get; set; }

      /// <summary>
      /// Get/set task parameters object
      /// </summary>
      ITaskParameters TaskParameters { get; set; }

      /// <summary>
      /// Test TDL expression for validity
      /// </summary>
      bool IsValid();

      /// <summary>
      /// Get/set enabled flag
      /// </summary>
      bool Enabled { get; set; }

      /// <summary>
      /// Get ISchedule object from parsing TDL
      /// </summary>
      /// <returns></returns>
      ISchedule GetSchedule();

      /// <summary>
      /// Get modified flag of entity
      /// </summary>
      bool IsModified();

      /// <summary>
      /// Refresh data from backend store
      /// </summary>
      void Refresh();

      /// <summary>
      /// Commit changes to backend store
      /// </summary>
      void Commit();
   }

   public enum TaskLoadMethod { AssemblyFile, AssemblyByName }

   public interface ITaskCollection : IDisposable, IEnumerable<ITaskDefinition> {
      /// <summary>
      /// Get task instance by name
      /// </summary>
      /// <param name="Name">Task name</param>
      /// <returns>ITaskDefinition object</returns>
      ITaskDefinition this[string Name] { get; }

      /// <summary>
      /// Get count of tasks
      /// </summary>
      int Count { get; }

      /// <summary>
      /// Check if a task exists
      /// </summary>
      /// <param name="name">Task name</param>
      /// <returns>True if exists</returns>
      bool Exists(string Name);

      /// <summary>
      /// Register a task
      /// </summary>
      /// <param name="TaskRegRef">TaskReg object</param>
      ITaskDefinition Add(string Name, string Description, TaskLoadMethod LoadMethod, string AssemblyName, string TypeName);

      /// <summary>
      /// Rename task
      /// </summary>
      /// <param name="OldName">Old task name</param>
      /// <param name="NewName">New task name</param>
      void Rename(string OldName, string NewName);

      /// <summary>
      /// Unregister a task
      /// </summary>
      /// <param name="ID">Task ID</param>
      void Delete(string Name);
   }

   public interface ITaskDefinition : IDisposable {
      /// <summary>
      /// Get task name
      /// </summary>
      string Name { get; }

      /// <summary>
      /// Get/set task description
      /// </summary>
      string Description { get; set; }

      /// <summary>
      /// Method used to load the ITask type
      /// </summary>
      TaskLoadMethod LoadMethod { get; set; }

      /// <summary>
      /// Get/set associated assembly name
      /// </summary>
      string AssemblyName { get; set; }

      /// <summary>
      /// Get/set associated type name
      /// </summary>
      string TypeName { get; set; }

      /// <summary>
      /// Get modified flag of entity
      /// </summary>
      /// <returns></returns>
      bool IsModified();

      /// <summary>
      /// Refresh data from backend store
      /// </summary>
      void Refresh();

      /// <summary>
      /// Commit changs to backend store
      /// </summary>
      void Commit();
   }

}
