﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Text;

namespace Expl.Itinerary {
   public class MemoryContextProvider {
      protected readonly DataTable _ScheduleTable, _JobTable, _TaskTable;

      public MemoryContextProvider() {
         // Define Schedule table
         {
            _ScheduleTable = new DataTable();
            DataColumn SchedulePK = _ScheduleTable.Columns.Add("Name", typeof(string));
            _ScheduleTable.PrimaryKey = new DataColumn[] { SchedulePK };
            DataColumn DescriptionColumn = _ScheduleTable.Columns.Add("Description", typeof(string));
            DescriptionColumn.AllowDBNull = false;
            DataColumn TDLColumn = _ScheduleTable.Columns.Add("TDL", typeof(string));
            TDLColumn.AllowDBNull = false;
         }

         // Define Job table
         {
            _JobTable = new DataTable();
            DataColumn JobPK = _JobTable.Columns.Add("ID", typeof(Guid));
            _JobTable.PrimaryKey = new DataColumn[] { JobPK };
            DataColumn NameColumn = _JobTable.Columns.Add("Name", typeof(string));
            NameColumn.AllowDBNull = false;
            DataColumn DescriptionColumn = _JobTable.Columns.Add("Description", typeof(string));
            DescriptionColumn.AllowDBNull = false;
            DataColumn TDLColumn = _JobTable.Columns.Add("TDL", typeof(string));
            TDLColumn.AllowDBNull = false;
            DataColumn TaskNameColumn = _JobTable.Columns.Add("TaskName", typeof(string));
            TaskNameColumn.AllowDBNull = false;
            DataColumn EnabledColumn = _JobTable.Columns.Add("Enabled", typeof(bool));
            EnabledColumn.AllowDBNull = false;
            DataColumn TaskParametersColumn = _JobTable.Columns.Add("TaskParameters", typeof(ITaskParameters));
            TaskParametersColumn.AllowDBNull = true;
         }

         // Define Task table
         {
            _TaskTable = new DataTable();
            DataColumn TaskPK = _TaskTable.Columns.Add("Name", typeof(string));
            _TaskTable.PrimaryKey = new DataColumn[] { TaskPK };
            DataColumn DescriptionColumn = _TaskTable.Columns.Add("Description", typeof(string));
            DescriptionColumn.AllowDBNull = false;
            DataColumn LoadMethodColumn = _TaskTable.Columns.Add("LoadMethod", typeof(TaskLoadMethod));
            LoadMethodColumn.AllowDBNull = false;
            DataColumn AssemblyNameColumn = _TaskTable.Columns.Add("AssemblyName", typeof(string));
            AssemblyNameColumn.AllowDBNull = false;
            DataColumn TypeNameColumn = _TaskTable.Columns.Add("TypeName", typeof(string));
            TypeNameColumn.AllowDBNull = false;
         }

         InitializeDefaultSchedules();
         InitializeDefaultTasks();
      }

      public DataTable ScheduleTable {
         get { return _ScheduleTable; }
      }

      public DataTable JobTable {
         get { return _JobTable; }
      }

      public DataTable TaskTable {
         get { return _TaskTable; }
      }

      private void InitializeDefaultSchedules() {
         foreach (DefaultScheduleItem Item in Defaults.DefaultSchedules) {
            DataRow Row = _ScheduleTable.NewRow();
            Row["Name"] = Item.Name;
            Row["Description"] = Item.Description;
            Row["TDL"] = Item.TDL;
            _ScheduleTable.Rows.Add(Row);
         }
      }

      private void InitializeDefaultTasks() {
         DataRow Row = _TaskTable.NewRow();
         Row["Name"] = "NOP";
         Row["Description"] = "No operation";
         Row["LoadMethod"] = TaskLoadMethod.AssemblyByName;
         Row["AssemblyName"] = typeof(Expl.Itinerary.Tasks.NOPTask).Assembly.ToString();
         Row["TypeName"] = typeof(Expl.Itinerary.Tasks.NOPTask).ToString();
         _TaskTable.Rows.Add(Row);
      }
   }

   public class MemoryItinerary : IItinerary {
      protected readonly MemoryContextProvider _Context;
      protected readonly MemoryScheduleCollection _Schedules;
      protected readonly MemoryJobCollection _Jobs;
      protected readonly MemoryTaskCollection _Tasks;

      public MemoryItinerary() : this(new MemoryContextProvider()) { }

      public MemoryItinerary(MemoryContextProvider Context) {
         _Context = Context;
         _Schedules = new MemoryScheduleCollection(this);
         _Jobs = new MemoryJobCollection(this);
         _Tasks = new MemoryTaskCollection(this);
      }

      public void Dispose() {
         _Tasks.Dispose();
         _Jobs.Dispose();
         _Schedules.Dispose();
         GC.SuppressFinalize(this);
      }

      public MemoryContextProvider Context {
         get { return _Context; }
      }

      public IScheduleCollection Schedules {
         get { return _Schedules; }
      }

      public IJobCollection Jobs {
         get { return _Jobs; }
      }

      public ITaskCollection Tasks {
         get { return _Tasks; }
      }
   }

   public class MemoryScheduleCollection : IScheduleCollection {
      private MemoryItinerary _Itinerary;

      public MemoryScheduleCollection(MemoryItinerary Itinerary) {
         _Itinerary = Itinerary;
      }

      public void Dispose() {
         _Itinerary = null;
         GC.SuppressFinalize(this);
      }

      public IScheduleDefinition this[string Name] {
         get {
            return new MemoryScheduleDefinition(_Itinerary, Name);
         }
      }

      public bool Exists(string Name) {
         return !(_Itinerary.Context.ScheduleTable.Rows.Find(Name) == null);
      }

      public int Count {
         get {
            return _Itinerary.Context.ScheduleTable.Rows.Count;
         }
      }

      public IScheduleDefinition Add(string Name, string Description, string TDL) {
         DataRow Row = _Itinerary.Context.ScheduleTable.NewRow();
         Row["Name"] = Name;
         Row["Description"] = Description;
         Row["TDL"] = TDL;
         _Itinerary.Context.ScheduleTable.Rows.Add(Row);
         return this[Name];
      }

      public void Rename(string OldName, string NewName) {
         DataRow Row = _Itinerary.Context.ScheduleTable.Rows.Find(OldName);
         Row["Name"] = NewName;
      }

      public void Delete(string Name) {
         _Itinerary.Context.ScheduleTable.Rows.Remove(_Itinerary.Context.ScheduleTable.Rows.Find(Name));
      }

      IEnumerator IEnumerable.GetEnumerator() {
         return (IEnumerator)this.GetEnumerator();
      }

      public IEnumerator<IScheduleDefinition> GetEnumerator() {
         foreach (DataRow Row in _Itinerary.Context.ScheduleTable.Rows) {
            yield return this[(string)Row["Name"]];
         }
         yield break;
      }
   }

   public class MemoryScheduleDefinition : IScheduleDefinition {
      private MemoryItinerary _Itinerary;
      private string _Name;
      private string _Description;
      private string _TDL;
      private bool _ModifiedFlag = false;
      private bool _LoadedFlag = false;

      public MemoryScheduleDefinition(MemoryItinerary Itinerary, string Name) {
         _Itinerary = Itinerary;
         _Name = Name;
      }

      public string Name {
         get { return _Name; }
      }

      public string Description {
         get {
            if (!_LoadedFlag) Refresh();
            return _Description;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _Description = value;
         }
      }

      public string TDL {
         get {
            if (!_LoadedFlag) Refresh();
            return _TDL;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TDL = value;
         }
      }

      public bool IsValid() {
         return GetSchedule() != null;
      }

      public ISchedule GetSchedule() {
         try {
            return TDLParser.Parse(TDL, new MemoryItinerary(_Itinerary.Context));
         }
         catch {
            return null;
         }
      }

      public bool IsModified() {
         return _ModifiedFlag;
      }

      public void Refresh() {
         DataRow Row = _Itinerary.Context.ScheduleTable.Rows.Find(_Name);
         _Description = (string)Row["Description"];
         _TDL = (string)Row["TDL"];
         _ModifiedFlag = false;
         _LoadedFlag = true;
      }

      public void Commit() {
         if (_ModifiedFlag) {
            DataRow Row = _Itinerary.Context.ScheduleTable.Rows.Find(_Name);
            Row["Description"] = Description;
            Row["TDL"] = TDL;
            _ModifiedFlag = false;
         }
      }

      public void Dispose() {
         _Itinerary = null;
         _Name = null;
         _Description = null;
         _TDL = null;
      }

      public override string ToString() {
         return _Name;
      }
   }

   public class MemoryJobCollection : IJobCollection {
      private MemoryItinerary _Itinerary;

      public event JobEventHandler JobAdded;
      public event JobEventHandler JobDeleted;
      public event JobEventHandler JobChanged;

      public MemoryJobCollection(MemoryItinerary Itinerary) {
         _Itinerary = Itinerary;
      }

      public IJobDefinition this[Guid ID] {
         get { 
            return new MemoryJobDefinition(_Itinerary, ID);
         }
      }

      public int Count {
         get {
            return _Itinerary.Context.JobTable.Rows.Count;
         }
      }

      public bool Exists(Guid ID) {
         DataRow Row = _Itinerary.Context.JobTable.Rows.Find(ID);
         return !(Row == null);
      }

      public IJobDefinition Add(string Name, string Description, string TDL, string TaskName, ITaskParameters TaskParameters, bool Enabled) {
         DataRow Row = _Itinerary.Context.JobTable.NewRow();
         Guid ID = Guid.NewGuid();
         Row["ID"] = ID;
         Row["Name"] = Name;
         Row["Description"] = Description;
         Row["TDL"] = TDL;
         Row["TaskName"] = TaskName;
         Row["TaskParameters"] = TaskParameters;
         Row["Enabled"] = Enabled;
         _Itinerary.Context.JobTable.Rows.Add(Row);
         IJobDefinition Job1 = this[ID];

         // Fire JobAdded event
         if (JobAdded != null) JobAdded(this, new JobEventArgs(Job1));

         return Job1;
      }

      public void Delete(Guid ID) {
         IJobDefinition Job1 = this[ID];
         Job1.Refresh();
         _Itinerary.Context.JobTable.Rows.Remove(_Itinerary.Context.JobTable.Rows.Find(ID));
         if (JobDeleted != null) JobDeleted(this, new JobEventArgs(Job1));
      }

      IEnumerator IEnumerable.GetEnumerator() {
         return (IEnumerator)this.GetEnumerator();
      }

      public IEnumerator<IJobDefinition> GetEnumerator() {
         foreach (DataRow Row in _Itinerary.Context.JobTable.Rows) {
            yield return this[(Guid)Row["ID"]];
         }
         yield break;
      }

      public void OnChanged(IJobDefinition JobEntity) {
         if (JobChanged != null) JobChanged(this, new JobEventArgs(JobEntity));
      }

      public void Dispose() {
         _Itinerary = null;
      }
   }

   public class MemoryJobDefinition : IJobDefinition {
      private MemoryItinerary _Itinerary;
      private Guid _ID;
      private string _Name;
      private string _Description;
      private string _TDL;
      private string _TaskName;
      private ITaskParameters _TaskParameters;
      private bool _Enabled;
      private bool _ModifiedFlag = false;
      private bool _LoadedFlag = false;

      public MemoryJobDefinition(MemoryItinerary Itinerary, Guid ID) {
         _Itinerary = Itinerary;
         _ID = ID;
      }

      public Guid ID {
         get { return _ID; }
      }

      public string Name {
         get {
            if (!_LoadedFlag) Refresh();
            return _Name;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _Name = value;
         }
      }

      public string Description {
         get {
            if (!_LoadedFlag) Refresh();
            return _Description;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _Description = value;
         }
      }

      public string TDL {
         get {
            if (!_LoadedFlag) Refresh();
            return _TDL;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TDL = value;
         }
      }

      public string TaskName {
         get {
            if (!_LoadedFlag) Refresh();
            return _TaskName;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TaskName = value;
         }
      }

      public ITaskParameters TaskParameters {
         get {
            if (!_LoadedFlag) Refresh();
            return _TaskParameters;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TaskParameters = value;
         }
      }

      public bool Enabled {
         get {
            if (!_LoadedFlag) Refresh();
            return _Enabled;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _Enabled = value;
         }
      }

      public bool IsValid() {
         return GetSchedule() != null;
      }

      public ISchedule GetSchedule() {
         try {
            return TDLParser.Parse(TDL, new MemoryItinerary(_Itinerary.Context));
         }
         catch {
            return null;
         }
      }

      public bool IsModified() {
         return _ModifiedFlag;
      }

      public void Refresh() {
         DataRow Row = _Itinerary.Context.JobTable.Rows.Find(_ID);
         _Name = (string)Row["Name"];
         _Description = (string)Row["Description"];
         _TDL = (string)Row["TDL"];
         _TaskName = (string)Row["TaskName"];
         _TaskParameters = Row["TaskParameters"] as ITaskParameters;
         _Enabled = (bool)Row["Enabled"];
         _ModifiedFlag = false;
         _LoadedFlag = true;
      }

      public void Commit() {
         if (_ModifiedFlag) {
            DataRow Row = _Itinerary.Context.JobTable.Rows.Find(_ID);
            Row["Name"] = Name;
            Row["Description"] = Description;
            Row["TDL"] = TDL;
            Row["TaskName"] = TaskName;
            Row["TaskParameters"] = TaskParameters;
            Row["Enabled"] = Enabled;
            ((MemoryJobCollection)_Itinerary.Jobs).OnChanged(this);
            _ModifiedFlag = false;
         }
      }

      public void Dispose() {
         _Itinerary = null;
         _Name = null;
         _Description = null;
         _TDL = null;
         _TaskName = null;
      }

      public override string ToString() {
         return Name;
      }
   }

   public class MemoryTaskCollection : ITaskCollection {
      private MemoryItinerary _Itinerary;

      public MemoryTaskCollection(MemoryItinerary Itinerary) {
         _Itinerary = Itinerary;
      }

      public ITaskDefinition this[string Name] {
         get {
            return new MemoryTaskDefinition(_Itinerary, Name);
         }
      }

      public int Count {
         get {
            return _Itinerary.Context.TaskTable.Rows.Count;
         }
      }

      public bool Exists(string Name) {
         DataRow Row = _Itinerary.Context.TaskTable.Rows.Find(Name);
         return !(Row == null);
      }

      public ITaskDefinition Add(string Name, string Description, TaskLoadMethod LoadMethod, string AssemblyName, string TypeName) {
         DataRow Row = _Itinerary.Context.TaskTable.NewRow();
         Row["Name"] = Name;
         Row["Description"] = Description;
         Row["LoadMethod"] = LoadMethod;
         Row["AssemblyName"] = AssemblyName;
         Row["TypeName"] = TypeName;
         _Itinerary.Context.TaskTable.Rows.Add(Row);
         return this[Name];
      }

      public void Rename(string OldName, string NewName) {
         DataRow Row = _Itinerary.Context.TaskTable.Rows.Find(OldName);
         Row["Name"] = NewName;
      }

      public void Delete(string Name) {
         _Itinerary.Context.TaskTable.Rows.Remove(_Itinerary.Context.TaskTable.Rows.Find(Name));
      }

      IEnumerator IEnumerable.GetEnumerator() {
         return (IEnumerator)this.GetEnumerator();
      }

      public IEnumerator<ITaskDefinition> GetEnumerator() {
         foreach (DataRow Row in _Itinerary.Context.TaskTable.Rows) {
            yield return this[(string)Row["Name"]];
         }
         yield break;
      }

      public void Dispose() {
         _Itinerary = null;
      }
   }

   public class MemoryTaskDefinition : ITaskDefinition {
      private MemoryItinerary _Itinerary;
      private string _Name;
      private string _Description;
      private TaskLoadMethod _LoadMethod;
      private string _AssemblyName;
      private string _TypeName;
      private bool _ModifiedFlag = false;
      private bool _LoadedFlag = false;

      public MemoryTaskDefinition(MemoryItinerary Itinerary, string Name) {
         _Itinerary = Itinerary;
         _Name = Name;
      }

      public string Name {
         get { return _Name; }
      }

      public string Description {
         get {
            if (!_LoadedFlag) Refresh();
            return _Description;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _Description = value;
         }
      }

      public TaskLoadMethod LoadMethod {
         get {
            if (!_LoadedFlag) Refresh();
            return _LoadMethod;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _LoadMethod = value;
         }
      }

      public string AssemblyName {
         get {
            if (!_LoadedFlag) Refresh();
            return _AssemblyName;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _AssemblyName = value;
         }
      }

      public string TypeName {
         get {
            if (!_LoadedFlag) Refresh();
            return _TypeName;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TypeName = value;
         }
      }

      public bool IsModified() {
         return _ModifiedFlag;
      }

      public override string ToString() {
         return this.Name;
      }

      public void Refresh() {
         DataRow Row = _Itinerary.Context.TaskTable.Rows.Find(_Name);
         _Description = (string)Row["Description"];
         _LoadMethod = (TaskLoadMethod)Row["LoadMethod"];
         _AssemblyName = (string)Row["AssemblyName"];
         _TypeName = (string)Row["TypeName"];
         _LoadedFlag = true;
         _ModifiedFlag = false;
      }

      public void Commit() {
         if (_ModifiedFlag) {
            DataRow Row = _Itinerary.Context.TaskTable.Rows.Find(_Name);
            Row["Description"] = Description;
            Row["AssemblyName"] = AssemblyName;
            Row["TypeName"] = TypeName;
            _ModifiedFlag = false;
         }
      }

      public void Dispose() {
         _Itinerary = null;
         _Name = null;
         _Description = null;
         _AssemblyName = null;
         _TypeName = null;
      }
   }
}
