using System;
using System.Collections;
using System.Collections.Generic;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Text;
using Expl.Auxiliary;

namespace Expl.Itinerary {
   /// <summary>
   /// Database session context object
   /// </summary>
   public class SqlItineraryContext : IDisposable {
      protected string _ConnString = null;
      protected SqlConnection _Conn = null;

      public SqlItineraryContext() { }

      public SqlItineraryContext(string ConnectionString) {
         _ConnString = ConnectionString;
      }

      /// <summary>
      /// Get connection string
      /// </summary>
      public string ConnString {
         get {
            if (_ConnString == null) {
               _ConnString = ConfigurationManager.ConnectionStrings["Itinerary"].ConnectionString;
               if (_ConnString == null) throw new Exception("Itinerary database connection string not found in app.config!");
            }
            return _ConnString;
         }
      }

      /// <summary>
      /// Get SqlConnection object
      /// </summary>
      public SqlConnection Conn {
         get {
            if (_Conn == null) {
               _Conn = new SqlConnection(ConnString);
               _Conn.Open();
            }
            return _Conn;
         }
      }

      public void InitializeDatabase() {
         // TO DO: Better way to split by "go" statement?
         foreach (string CmdString in Resource.SqlBuildDatabaseScript.Split(new string[] {"go" + Environment.NewLine}, StringSplitOptions.None)) {
            using (SqlCommand Cmd = new SqlCommand(CmdString, Conn)) {
               Cmd.ExecuteNonQuery();
            }
         }
      }

      public void Dispose() {
         _ConnString = null;
         _Conn = null;
      }
   }

   /// <summary>
   /// Implementation of IStateProvider to use SQL database backend
   /// </summary>
   public class SqlItinerary : IItinerary {
      protected readonly SqlItineraryContext _Context;
      protected readonly SqlScheduleCollection _Schedules;
      protected readonly SqlTaskCollection _Tasks;
      protected readonly SqlJobCollection _Jobs;

      public SqlItinerary() : this(new SqlItineraryContext()) { }

      public SqlItinerary(SqlItineraryContext Context) {
         _Context = Context;
         _Schedules = new SqlScheduleCollection(this);
         _Jobs = new SqlJobCollection(this);
         _Tasks = new SqlTaskCollection(this); 
      }

      public void Dispose() {
         _Tasks.Dispose();
         _Jobs.Dispose();
         _Schedules.Dispose();
         _Context.Dispose();
         GC.SuppressFinalize(this);
      }

      public SqlItineraryContext Context {
         get { return _Context; }
      }

      public IScheduleCollection Schedules {
         get { return _Schedules; }
      }

      public IJobCollection Jobs {
         get { return _Jobs; }
      }

      public ITaskCollection Tasks {
         get { return _Tasks; }
      }
   }

   /// <summary>
   /// SQL database implementation of IScheduleCollection
   /// </summary>
   public class SqlScheduleCollection : IScheduleCollection {
      private SqlItinerary _Itinerary;

      public SqlScheduleCollection(SqlItinerary Context) {
         _Itinerary = Context;
      }

      public void Dispose() {
         _Itinerary = null;
         GC.SuppressFinalize(this);
      }

      public IScheduleDefinition this[string Name] {
         get {
            // Check if Name exists
            using (SqlCommand Cmd = new SqlCommand("select count(*) from Schedule where [name]=@Name", _Itinerary.Context.Conn)) {
               Cmd.CommandType = CommandType.Text;
               Cmd.Parameters.AddWithValue("Name", Name);
               int RowCount = (int)Cmd.ExecuteScalar();
               if (RowCount == 0) throw new NotFoundException(string.Format("Schedule '{0}' not found", Name));
            }
            return new SqlScheduleDefinition(_Itinerary, Name);
         }
      }

      public bool Exists(string Name) {
         using (SqlCommand Cmd = new SqlCommand("select count(*) from Schedule where [name]=@Name", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("Name", Name);
            int count = (int)Cmd.ExecuteScalar();
            return count > 0;
         }
      }

      IEnumerator IEnumerable.GetEnumerator() {
         return this.GetEnumerator();
      }

      public IEnumerator<IScheduleDefinition> GetEnumerator() {
         using (SqlCommand Cmd = new SqlCommand("select [name] from Schedule", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            SqlDataAdapter adapter = new SqlDataAdapter(Cmd);
            DataTable dt = new DataTable();
            adapter.Fill(dt);
            foreach (DataRow row in dt.Rows) {
               yield return new SqlScheduleDefinition(_Itinerary, (string)row[0]);
            }
         }
         yield break;
      }

      public int Count {
         get {
            using (SqlCommand Cmd = new SqlCommand("select count(*) from Schedule", _Itinerary.Context.Conn)) {
               return (int)Cmd.ExecuteScalar();
            }
         }
      }

      public IScheduleDefinition Add(string Name, string Description, string TDL) {
         using (SqlCommand Cmd = new SqlCommand(
            "insert into Schedule ([name], [description], tdl) " +
            "values (@Name, @Description, @TDL)", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("Name", Name);
            Cmd.Parameters.AddWithValue("Description", Description);
            Cmd.Parameters.AddWithValue("TDL", TDL);
            int RowsAffected = Cmd.ExecuteNonQuery();
            if (RowsAffected == 0) throw new UpdateFailedException(string.Format("Error adding schedule: '{0}'", Name));
         }
         return this[Name];
      }

      public void Rename(string OldName, string NewName) {
         using (SqlCommand Cmd = new SqlCommand(
            "update Schedule set [name]=@NewName where [name]=@OldName", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("NewName", NewName);
            Cmd.Parameters.AddWithValue("OldName", OldName);
            int RowsAffected = Cmd.ExecuteNonQuery();
            if (RowsAffected == 0) throw new UpdateFailedException(string.Format("Error renaming schedule '{0}'", OldName));
         }
      }

      public void Delete(string Name) {
         using (SqlCommand Cmd = new SqlCommand(
            "delete from Schedule where [name] = @Name",
            _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("Name", Name);
            int RowsAffected = Cmd.ExecuteNonQuery();
            if (RowsAffected == 0) throw new UpdateFailedException(string.Format("Error deleting schedule: '{0}'", Name));
         }
      }
   }

   /// <summary>
   /// SQL database implementation of IScheduleDefinition
   /// </summary>
   public class SqlScheduleDefinition : IScheduleDefinition {
      private SqlItinerary _Itinerary;
      private string _Name, _Description, _TDL;
      private bool _ModifiedFlag = false;
      private bool _LoadedFlag = false;

      public SqlScheduleDefinition(SqlItinerary Itinerary, string Name) {
         _Itinerary = Itinerary;
         _Name = Name;
      }

      public string Name {
         get { return _Name; }
      }

      public string Description {
         get {
            if (!_LoadedFlag) Refresh();
            return _Description;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _Description = value;
         }
      }

      public string TDL {
         get {
            if (!_LoadedFlag) Refresh();
            return _TDL;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TDL = value;
         }
      }

      public bool IsValid() {
         return GetSchedule() != null;
      }

      public ISchedule GetSchedule() {
         try {
            return TDLParser.Parse(TDL, new SqlItinerary(_Itinerary.Context));
         }
         catch {
            return null;
         }
      }

      public bool IsModified() {
         return _ModifiedFlag;
      }

      public void Dispose() {
         _Itinerary = null;
         _Name = null;
         _Description = null;
         _TDL = null;
      }

      public override string ToString() {
         return this.Name;
      }

      public void Refresh() {
         using (SqlCommand Cmd = new SqlCommand("select top 1 [description], tdl from Schedule where [name] = @Name", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("Name", this.Name);
            SqlDataReader rdr = Cmd.ExecuteReader();
            try {
               if (rdr.HasRows) {
                  rdr.Read();
                  _Description = (string)rdr[0];
                  _TDL = (string)rdr[1];
               }
            }
            finally {
               rdr.Close();
            }
         }
         _LoadedFlag = true;
         _ModifiedFlag = false;
      }

      public void Commit() {
         if (_ModifiedFlag) {
            using (SqlCommand Cmd = new SqlCommand(
               "update Schedule set [description]=@Description, [TDL]=@TDL where @Name=[Name]", _Itinerary.Context.Conn)) {
               Cmd.CommandType = CommandType.Text;
               Cmd.Parameters.AddWithValue("Description", this.Description);
               Cmd.Parameters.AddWithValue("TDL", this.TDL);
               Cmd.Parameters.AddWithValue("Name", this.Name);
               int RowsAffected = Cmd.ExecuteNonQuery();
               if (RowsAffected == 0) throw new UpdateFailedException(string.Format("Error updating schedule '{0}'", Name));
            }
            _ModifiedFlag = false;
         }
      }
   }

   /// <summary>
   /// SQL database implementation of IHighTowerJobCollection
   /// </summary>
   public class SqlJobCollection : IJobCollection {
      private SqlItinerary _Itinerary;

      public event JobEventHandler JobAdded;
      public event JobEventHandler JobDeleted;
      public event JobEventHandler JobChanged;

      public SqlJobCollection(SqlItinerary Itinerary) {
         _Itinerary = Itinerary;
      }

      public void Dispose() {
         _Itinerary = null;
      }

      public IJobDefinition this[Guid ID] {
         get { return new SqlJobDefinition(_Itinerary, ID); }
      }

      public bool Exists(Guid ID) {
         using (SqlCommand Cmd = new SqlCommand("select count(*) from Job where id=@ID", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("ID", ID);
            int count = (int)Cmd.ExecuteScalar();
            return count > 0;
         }
      }

      IEnumerator IEnumerable.GetEnumerator() {
         return this.GetEnumerator();
      }

      public IEnumerator<IJobDefinition> GetEnumerator() {
         using (SqlCommand Cmd = new SqlCommand("select id from Job", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            SqlDataAdapter Adapter = new SqlDataAdapter(Cmd);
            DataTable dt = new DataTable();
            Adapter.Fill(dt);
            foreach (DataRow row in dt.Rows) {
               yield return new SqlJobDefinition(_Itinerary, (Guid)row[0]);
            }
            yield break;
         }
      }

      public int Count {
         get {
            using (SqlCommand Cmd = new SqlCommand("select count(*) from Job", _Itinerary.Context.Conn)) {
               Cmd.CommandType = CommandType.Text;
               return (int)Cmd.ExecuteScalar();
            }
         }
      }

      public IJobDefinition Add(string Name, string Description, string TDL, string TaskName, ITaskParameters TaskParameters, bool Enabled) {
         Guid ID = Guid.NewGuid();
         using (SqlCommand Cmd = new SqlCommand("insert into Job (id, [name], [description], tdl, task_name, task_parameters) values (@ID, @Name, @Description, @TDL, @TaskName, @TaskParameters)", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("ID", ID);
            Cmd.Parameters.AddWithValue("Name", Name);
            Cmd.Parameters.AddWithValue("Description", Description);
            Cmd.Parameters.AddWithValue("TDL", TDL);
            Cmd.Parameters.AddWithValue("TaskName", TaskName);

            // Serialize TaskParameters
            SqlParameter ParametersParam = new SqlParameter("TaskParameters", SqlDbType.Image);
            ParametersParam.Value = DBNull.Value;
            //if (TaskParameters != null) ParametersParam.Value = Serialize.SerializeToXml(TaskParameters);
            if (TaskParameters != null) ParametersParam.Value = Serialize.BinarySerialize(TaskParameters);

            Cmd.Parameters.Add(ParametersParam);
            
            int affected = Cmd.ExecuteNonQuery();
            if (affected == 0) throw new UpdateFailedException(string.Format("Error adding job ID: '{0}'", ID.ToString()));
         }
         IJobDefinition Job1 = this[ID];

         // Fire JobAdded event
         if (JobAdded != null) JobAdded(this, new JobEventArgs(Job1));

         return Job1;
      }

      public void Delete(Guid ID) {
         IJobDefinition Job1 = null;
         if (JobDeleted != null) Job1 = this[ID];

         using (SqlCommand Cmd = new SqlCommand("delete from Job where id=@ID", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("ID", ID);
            int RowsAffected = Cmd.ExecuteNonQuery();
            if (RowsAffected == 0) throw new UpdateFailedException(string.Format("Error deleting job ID: '{0}'", ID.ToString()));
         }

         if (JobDeleted != null) JobDeleted(this, new JobEventArgs(Job1));
      }

      public void OnChanged(IJobDefinition JobEntity) {
         if (JobChanged != null) JobChanged(this, new JobEventArgs(JobEntity));
      }
   }

   public class SqlJobDefinition : IJobDefinition {
      private SqlItinerary _Itinerary;
      private Guid _ID;
      private string _Name;
      private string _Description;
      private string _TDL;
      private string _TaskName;
      private ITaskParameters _TaskParameters;
      private bool _EnabledFlag;
      private bool _ModifiedFlag = false;
      private bool _LoadedFlag = false;

      public SqlJobDefinition(SqlItinerary Itinerary, Guid ID) {
         _Itinerary = Itinerary;
         _ID = ID;
      }

      public Guid ID {
         get { return _ID; }
      }

      public string Name {
         get {
            if (!_LoadedFlag) Refresh();
            return _Name;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _Name = value;
         }
      }

      public string Description {
         get {
            if (!_LoadedFlag) Refresh();
            return _Description; 
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _Description = value;
         }
      }

      public string TDL {
         get {
            if (!_LoadedFlag) Refresh();
            return _TDL;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TDL = value;
         }
      }

      public string TaskName {
         get {
            if (!_LoadedFlag) Refresh();
            return _TaskName;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TaskName = value;
         }
      }

      public ITaskParameters TaskParameters {
         get {
            if (!_LoadedFlag) Refresh();
            return _TaskParameters;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TaskParameters = value;
         }
      }

      public bool Enabled {
         get {
            if (!_LoadedFlag) Refresh();
            return _EnabledFlag;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _EnabledFlag = value;
         }
      }

      public bool IsValid() {
         return GetSchedule() != null;
      }

      public ISchedule GetSchedule() {
         try {
            return TDLParser.Parse(TDL, new SqlItinerary(_Itinerary.Context));
         }
         catch {
            return null;
         }
      }

      public void Dispose() {
         _Name = null;
         _Description = null;
         _TDL = null;
         _TaskName = null;
      }

      public override string ToString() {
         return Name;
      }

      public bool IsModified() {
         return _ModifiedFlag;
      }

      public void Refresh() {
         using (SqlCommand Cmd = new SqlCommand(
            "select top 1 [name], [description], tdl, task_name, " +
            "task_parameters, enabled from Job where id=@ID", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("ID", _ID);
            SqlDataReader rdr = Cmd.ExecuteReader();
            
            try {
               if (rdr.HasRows) {
                  rdr.Read();
                  _Name = (string)rdr[0];
                  _Description = (string)rdr[1];
                  _TDL = (string)rdr[2];
                  _TaskName = (string)rdr[3];
                  _TaskParameters = rdr[4] is DBNull ? null : (ITaskParameters)Serialize.BinaryDeserialize((byte[])rdr[4]);
                  _EnabledFlag = (bool)rdr[5];
               }
            }
            finally {
               rdr.Close();
            }

         }
         _LoadedFlag = true;
         _ModifiedFlag = false;
      }

      public void Commit() {
         if (_ModifiedFlag) {
            using (SqlCommand Cmd = new SqlCommand(
               "update Job set [name]=@Name, [description]=@Description, " +
               "tdl=@TDL, task_name=@TaskName, task_parameters=@TaskParameters, " +
               "enabled=@Enabled where id=@ID", _Itinerary.Context.Conn)) {
               Cmd.CommandType = CommandType.Text;
               Cmd.Parameters.AddWithValue("ID", _ID);
               Cmd.Parameters.AddWithValue("Name", Name);
               Cmd.Parameters.AddWithValue("Description", Description);
               Cmd.Parameters.AddWithValue("TDL", TDL);
               Cmd.Parameters.AddWithValue("TaskName", TaskName);

               SqlParameter ParametersParam = new SqlParameter("TaskParameters", DBNull.Value);
               if (TaskParameters != null) ParametersParam.Value = Serialize.BinarySerialize(TaskParameters);

               Cmd.Parameters.AddWithValue("Enabled", Enabled);
               int RowsAffected = Cmd.ExecuteNonQuery();
               if (RowsAffected == 0) throw new UpdateFailedException(string.Format("Error updating job '{0}'", _ID.ToString()));
            }
            ((SqlJobCollection)_Itinerary.Jobs).OnChanged(this);
            _ModifiedFlag = false;
         }
      }
   }

   public class SqlTaskCollection : ITaskCollection {
      private SqlItinerary _Itinerary;

      public SqlTaskCollection(SqlItinerary Itinerary) {
         _Itinerary = Itinerary;
      }

      public void Dispose() {
         _Itinerary = null;
      }

      public ITaskDefinition this[string Name] {
         get { return new SqlTaskDefinition(_Itinerary, Name); }
      }

      public int Count {
         get {
            using (SqlCommand Cmd = new SqlCommand("select count(*) from Task", _Itinerary.Context.Conn)) {
               Cmd.CommandType = CommandType.Text;
               return (int)Cmd.ExecuteScalar();
            }
         }
      }

      public bool Exists(string Name) {
         using (SqlCommand Cmd = new SqlCommand("select count(*) from Task where [name]=@Name", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("Name", Name);
            int count = (int)Cmd.ExecuteScalar();
            return count > 0;
         }
      }


      public ITaskDefinition Add(string Name, string Description, TaskLoadMethod LoadMethod, string AssemblyName, string TypeName) {
         using (SqlCommand Cmd = new SqlCommand("insert into Task ([name], [description], load_method, assembly_name, [type_name]) values (@Name, @Description, @LoadMethod, @AssemblyName, @TypeName)", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("Name", Name);
            Cmd.Parameters.AddWithValue("Description", Description);
            Cmd.Parameters.AddWithValue("LoadMethod", (int)LoadMethod);
            Cmd.Parameters.AddWithValue("AssemblyName", AssemblyName);
            Cmd.Parameters.AddWithValue("TypeName", TypeName);
            int affected = Cmd.ExecuteNonQuery();
            if (affected == 0) throw new UpdateFailedException(string.Format("Error adding Task: '{0}'", Name));
         }
         return this[Name];
      }

      public void Rename(string OldName, string NewName) {
         using (SqlCommand Cmd = new SqlCommand(
            "update Task set [name]=@NewName where [name]=@OldName", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("NewName", NewName);
            Cmd.Parameters.AddWithValue("OldName", OldName);
            int RowsAffected = Cmd.ExecuteNonQuery();
            if (RowsAffected == 0) throw new UpdateFailedException(string.Format("Error renaming task '{0}'", OldName));
         }
      }

      public void Delete(string Name) {
         using (SqlCommand Cmd = new SqlCommand("delete from Task where [name]=@Name", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("Name", Name);
            int affected = Cmd.ExecuteNonQuery();
            if (affected == 0) throw new NotFoundException(string.Format("Error deleting Task: '{0}'", Name));
         }
      }

      IEnumerator IEnumerable.GetEnumerator() {
         return (IEnumerator)this.GetEnumerator();
      }

      public IEnumerator<ITaskDefinition> GetEnumerator() {
         using (SqlCommand Cmd = new SqlCommand("select [name] from Task order by [name]", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            SqlDataAdapter Adapter = new SqlDataAdapter(Cmd);
            DataTable dt = new DataTable();
            Adapter.Fill(dt);
            foreach (DataRow row in dt.Rows) {
               yield return new SqlTaskDefinition(_Itinerary, (string)row[0]);
            }
         }
         yield break;
      }
   }

   /// <summary>
   /// SQL database implementation of ITaskDefinition
   /// </summary>
   public class SqlTaskDefinition : ITaskDefinition {
      private SqlItinerary _Itinerary;
      private string _Name;
      private string _Description;
      private TaskLoadMethod _LoadMethod;
      private string _AssemblyName;
      private string _TypeName;
      private bool _ModifiedFlag = false;
      private bool _LoadedFlag = false;

      public SqlTaskDefinition(SqlItinerary Itinerary, string Name) {
         _Itinerary = Itinerary;
         _Name = Name;
      }

      public void Dispose() {
         _Itinerary = null;
         _Name = null;
         _Description = null;
         _AssemblyName = null;
         _TypeName = null;
      }

      public string Name {
         get { return _Name; }
      }

      public string Description {
         get {
            if (!_LoadedFlag) Refresh();
            return _Description;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _Description = value;
         }
      }

      public TaskLoadMethod LoadMethod {
         get {
            if (!_LoadedFlag) Refresh();
            return _LoadMethod;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _LoadMethod = value;
         }
      }

      public string AssemblyName {
         get {
            if (!_LoadedFlag) Refresh();
            return _AssemblyName;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _AssemblyName = value;
         }
      }

      public string TypeName {
         get {
            if (!_LoadedFlag) Refresh();
            return _TypeName;
         }
         set {
            if (!_LoadedFlag) Refresh();
            _ModifiedFlag = true;
            _TypeName = value;
         }
      }

      public bool IsModified() {
         return _ModifiedFlag;
      }

      public override string ToString() {
         return this.Name;
      }

      public void Refresh() {
         using (SqlCommand Cmd = new SqlCommand("select top 1 [description], load_method, assembly_name, [type_name] from Task where [name] = @Name", _Itinerary.Context.Conn)) {
            Cmd.CommandType = CommandType.Text;
            Cmd.Parameters.AddWithValue("Name", this.Name);
            SqlDataReader rdr = Cmd.ExecuteReader();
            try {
               if (rdr.HasRows) {
                  rdr.Read();
                  _Description = (string)rdr[0];
                  _LoadMethod = (TaskLoadMethod)rdr[1];
                  _AssemblyName = (string)rdr[2];
                  _TypeName = (string)rdr[3];
               }
            }
            finally {
               rdr.Close();
            }
         }
         _LoadedFlag = true;
         _ModifiedFlag = false;
      }

      public void Commit() {
         if (_ModifiedFlag) {
            using (SqlCommand Cmd = new SqlCommand(
               "update Task set [description]=@Description, load_method=@LoadMethod, assembly_name=@AssemblyName, [type_name]=@TypeName where @Name=[Name]", _Itinerary.Context.Conn)) {
               Cmd.CommandType = CommandType.Text;
               Cmd.Parameters.AddWithValue("Description", this.Description);
               Cmd.Parameters.AddWithValue("LoadMethod", this.LoadMethod);
               Cmd.Parameters.AddWithValue("AssemblyName", this.AssemblyName);
               Cmd.Parameters.AddWithValue("TypeName", this.TypeName);
               Cmd.Parameters.AddWithValue("Name", this.Name);
               int RowsAffected = Cmd.ExecuteNonQuery();
               if (RowsAffected == 0) throw new UpdateFailedException(string.Format("Error updating task '{0}'", this.Name));
            }
            _ModifiedFlag = false;
         }
      }
   }

}
