using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using Expl.Auxiliary;

namespace Expl.Itinerary {

   /// <summary>
   /// Intersecting schedules collection of two ISchedule objects where event times overlap.
   /// </summary>
   [Description("Intersection")]
   public class IntersectionSchedule : ICompositeSchedule {
      protected ISchedule _ScheduleA, _ScheduleB;

      /// <summary>
      /// Optimization threshold for iterating two schedules.
      /// If iterating more than SeekCountThreshold and no intersection has been found, 
      /// the iterator will be rebuilt starting at the next predicted match.
      /// </summary>
      protected const int SeekCountThreshold = 2;

      /// <summary>
      /// Default constructor
      /// <remarks>Uses Void schedules as parameters</remarks>
      /// </summary>
      public IntersectionSchedule()
         : this(new VoidSchedule(), new VoidSchedule()) { }

      /// <summary>
      /// Constructor
      /// </summary>
      /// <param name="ScheduleA">Schedule A</param>
      /// <param name="ScheduleB">Schedule B</param>
      public IntersectionSchedule(ISchedule ScheduleA, ISchedule ScheduleB) {
         _ScheduleA = ScheduleA;
         _ScheduleB = ScheduleB;
      }

      /// <summary>
      /// Constructor for schedule list
      /// <remarks>Will only enumerate the first two items from the list.
      /// If the list contains less than 2 items, default to Void schedules.</remarks>
      /// </summary>
      /// <param name="List">Enumerable list of ISchedule objects</param>
      public IntersectionSchedule(IEnumerable<ISchedule> List) {
         var Iter = List.GetEnumerator();
         if (Iter.MoveNext()) {
            _ScheduleA = Iter.Current;
            if (Iter.MoveNext())
               _ScheduleB = Iter.Current;
            else
               _ScheduleB = new VoidSchedule();
         }
         else {
            _ScheduleA = _ScheduleB = new VoidSchedule();
         }
      }

      public void Dispose() {
         _ScheduleA = null;
         _ScheduleB = null;
      }

      public override string ToString() {
         return string.Format("({0} & {1})", _ScheduleA.ToString(), _ScheduleB.ToString());
      }

      public IEnumerable<ISchedule> Schedules {
         get {
            yield return _ScheduleA;
            yield return _ScheduleB;
         }
      }

      public int MinSchedules {
         get { return 2; }
      }

      public int MaxSchedules {
         get { return 2; }
      }

      // Optimized for iterative access 2008-07-16
      // Optimized with SeekCount 2008-07-18
      public IEnumerable<TimedEvent> GetRange(DateTime RangeStart, DateTime RangeEnd) {
         yield break;
      }
   }
}
