using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Diagnostics;

namespace Expl.Itinerary {
   /// <summary>
   /// Create ITask instance from ITaskDefinition
   /// </summary>
   public static class TaskFactory {
      /// <summary>
      /// Create task object
      /// </summary>
      /// <param name="TaskEntity">Task definition</param>
      /// <returns>ITask object</returns>
      public static ITask CreateTask(ITaskDefinition TaskEntity) {
         Assembly a = LoadAssembly(TaskEntity);
         Type TaskType = a.GetType(TaskEntity.TypeName);
         if (TaskType == null) throw new Exception(string.Format("Unable to load class '{0}' from task '{1}'", TaskEntity.TypeName, TaskEntity.Name));

         // Veirfy TaskType implements ITask
         if (TaskType.GetInterface(typeof(ITask).ToString()) == null) throw new Exception(string.Format("Attempted to instantiate ITask class '{0}', but it does not implement ITask.", TaskEntity.TypeName));

         return (ITask)CreateInstance(a, TaskType, null);
      }

      /// <summary>
      /// Create a parameter entity object
      /// </summary>
      /// <param name="TaskEntity">Task definition</param>
      /// <returns>ITaskParameters object</returns>
      public static ITaskParameters CreateParameters(ITaskDefinition TaskEntity) {
         // TO DO: Get Task class metadata and find the ITaskParameters class name to instantiate
         Assembly a = LoadAssembly(TaskEntity);
         Type TaskType = a.GetType(TaskEntity.TypeName);
         if (TaskType == null) throw new Exception(string.Format("Unable to load class '{0}' from task '{1}'", TaskEntity.TypeName, TaskEntity.Name));
         object[] AttrArr = TaskType.GetCustomAttributes(typeof(ParametersTypeName), false);
         if (AttrArr.Length == 0) throw new Exception(string.Format("'ParametersTypeName' attribute not defined on class '{0}' from task '{1}'", TaskEntity.TypeName, TaskEntity.Name));

         ParametersTypeName Attr = (ParametersTypeName)AttrArr[0];
         Type ParametersType = a.GetType(Attr.TypeName);
         if (ParametersType == null) throw new ApplicationException(string.Format("Unable to load class '{0}'", Attr.TypeName));

         // Verify ParametersType implements ITaskParameters
         if (ParametersType.GetInterface(typeof(ITaskParameters).ToString()) == null) throw new ApplicationException(string.Format("Attempted to instantiate ITaskParameters class '{0}', but this type does not implement ITaskParameters.", ParametersType.ToString()));

         return (ITaskParameters)CreateInstance(a, ParametersType, null);
      }

      /// <summary>
      /// Load assembly referenced in ITaskDefinition
      /// </summary>
      /// <param name="TaskEntity"></param>
      /// <returns></returns>
      private static Assembly LoadAssembly(ITaskDefinition TaskEntity) {
         switch (TaskEntity.LoadMethod) {
         case TaskLoadMethod.AssemblyFile:
            return Assembly.LoadFrom(TaskEntity.AssemblyName);

         case TaskLoadMethod.AssemblyByName:
            return Assembly.Load(new AssemblyName(TaskEntity.AssemblyName));

         default:
            throw new Exception(string.Format("{0} found invalid TaskLoadMethodValue of '{1:d}'", Expl.Auxiliary.ECDiagnostics.ThisMethod, (int)TaskEntity.LoadMethod));
         }
      }

      /// <summary>
      /// Create instance of a loaded assembly's type
      /// TypeRef can be extraced from AssemblyRef.GetType()
      /// </summary>
      /// <param name="AssemblyRef">Assembly reference</param>
      /// <param name="TypeRef">Type to instantiate</param>
      /// <param name="CtorArguments">Constructor arguments, or null for none</param>
      /// <returns>Instantiated object reference</returns>
      private static object CreateInstance(Assembly AssemblyRef, Type TypeRef, object[] CtorArguments) {
         object o;
         if (CtorArguments == null) {
            o = TypeRef.GetConstructor(new Type[] { }).Invoke(new object[] { });
         }
         else {
            Type[] ArgTypes = new Type[CtorArguments.Length];
            for (int i = 0; i < CtorArguments.Length; i++) {
               ArgTypes[i] = CtorArguments[i].GetType();
            }
            o = TypeRef.GetConstructor(ArgTypes).Invoke(CtorArguments);
         }
         return o;
      }
   }
}
